//
//  P4WorkspaceDefaults.m
//  Perforce
//
//  Created by Adam Czubernat on 13/12/2013.
//  Copyright (c) 2013 Perforce Software, Inc. All rights reserved.
//

#import "P4WorkspaceDefaults.h"
#import "P4Workspace.h"
#import "P4Defaults.h"

NSString * const P4WorkspaceDefaultsChangedNotification = @"P4WorkspaceDefaultsChangedNotification";
NSString * const P4WorkspaceDefaultsTagFilterNotification = @"P4WorkspaceDefaultsTagFilterNotification";
NSString * const P4WorkspaceDefaultsProjectsDirectoryName = @"Projects";

static P4WorkspaceDefaults *sharedInstance;

@interface P4WorkspaceDefaults () {
	NSString *workspace;
}
@property (nonatomic, retain) NSDictionary *mapping;
- (id)initWithWorkspace:(NSString *)workspace;
- (NSURL *)projectsDirectoryURL;
- (void)updateProjectsDirectory;
- (void)updateSymbolicLinks:(NSArray *)created deleted:(NSArray *)deleted;
- (void)updateFinderSidebarItems:(NSArray *)created deleted:(NSArray *)deleted PS_DEPRECATED; // Optional making finder items for every mapped folder
@end

@implementation P4WorkspaceDefaults
@synthesize favoriteFolders, favoriteTags, favoriteNames, filteredTags;
@synthesize iconView;
@synthesize mapping;

#pragma mark - PSDefaults override

+ (P4WorkspaceDefaults *)sharedInstance {
	return sharedInstance;
}

- (NSString *)keyPath {
	NSAssert(workspace, @"P4WorkspaceDefaults key missing a workspace");
	return [NSString stringWithFormat:@"Workspaces.%@", workspace];
}

#pragma mark - Public

- (id)initWithWorkspace:(NSString *)workspaceName {
	workspace = [workspaceName stringByReplacingOccurrencesOfString:@"." withString:@"_"];
	NSAssert(workspace, @"P4WorkspaceDefaults missing a workspace");
	if (self = [super init]) {
		[self updateProjectsDirectory];
		[self updateMapping];
	}
	return self;
}

- (NSURL *)projectsDirectoryURL {
	static NSURL *directoryURL;
	
	if (!directoryURL) {
//		NSArray *URLs = [[NSFileManager defaultManager]
//						 URLsForDirectory:NSDocumentDirectory inDomains:NSUserDomainMask];
//		NSURL *parentURL = [URLs firstObject];
		NSURL *parentURL = [NSURL fileURLWithPath:NSHomeDirectory() isDirectory:YES];
		NSString *directoryName = P4WorkspaceDefaultsProjectsDirectoryName;
		directoryURL = [parentURL URLByAppendingPathComponent:directoryName isDirectory:YES];
	}

	if (![directoryURL checkResourceIsReachableAndReturnError:NULL]) {
		[[NSFileManager defaultManager]
		 createDirectoryAtURL:directoryURL
		 withIntermediateDirectories:NO
		 attributes:0
		 error:NULL];
	}
	return directoryURL;
}

- (void)updateProjectsDirectory {
	NSURL *directoryURL = [self projectsDirectoryURL];
	
	// Check finder items
	NSArray *finderItems = [[NSWorkspace sharedWorkspace] finderSidebarItems];

	if (![finderItems containsObject:directoryURL]) {
		[[NSWorkspace sharedWorkspace] insertFinderSidebarItem:directoryURL afterItem:nil];
	}
}

- (void)updateSymbolicLinks:(NSArray *)created deleted:(NSArray *)deleted {
	
	NSMutableArray *symlinks = [NSMutableArray array];
	NSFileManager *filemanager = [NSFileManager defaultManager];

	NSURL *projectsURL = [self projectsDirectoryURL];
	NSArray *projectsContents = [filemanager
								 contentsOfDirectoryAtURL:projectsURL
								 includingPropertiesForKeys:
								 @[ NSURLIsDirectoryKey,
									NSURLNameKey,
									NSURLIsSymbolicLinkKey, ]
								 options:NSDirectoryEnumerationSkipsHiddenFiles
								 error:NULL];
	// Enumerate directory
	for (NSURL *fileUrl in projectsContents) {
		NSNumber *isSymlink = nil;
		[fileUrl getResourceValue:&isSymlink forKey:NSURLIsSymbolicLinkKey error:NULL];
		if (isSymlink.boolValue) {
			NSString *dest = [filemanager destinationOfSymbolicLinkAtPath:fileUrl.path error:NULL];
			[symlinks addObject:dest];
			// Remove deleted symlinks
			if ([deleted containsObject:dest])
				[filemanager removeItemAtURL:fileUrl error:NULL];
		}
	}
	
	for (NSString *path in created) {
		// Create symlink
		if (![symlinks containsObject:path]) {
			NSString *name = [path lastPathComponent];
			
			// Check for duplicate name
			NSString *projectsPath = projectsURL.path;
			NSString *symlinkPath = [projectsPath stringByAppendingPathComponent:name];
			for (NSInteger i = 1; [filemanager fileExistsAtPath:symlinkPath]; i++)
				symlinkPath = [projectsPath stringByAppendingPathComponent:
							   [NSString stringWithFormat:@"%@ %ld", name, i]];

			/* Commented because it's problematic for map-unmap when deleting from remote */
//			// Create folder if doesn't exist
//			if ([path hasSuffix:@"/"] &&
//				[path isSubpath:[P4Workspace sharedInstance].root] &&
//				![filemanager fileExistsAtPath:path]) {
//				[filemanager createDirectoryAtPath:path withIntermediateDirectories:YES
//										attributes:nil error:NULL];
//			}
			
			[filemanager
			 createSymbolicLinkAtURL:[NSURL fileURLWithPath:symlinkPath isDirectory:NO]
			 withDestinationURL:[NSURL fileURLWithPath:path isDirectory:YES]
			 error:NULL];
		}
	}
}

// Optional making finder items for every mapped folder
- (void)updateFinderSidebarItems:(NSArray *)created deleted:(NSArray *)deleted {
	
	NSArray *mappingPaths = self.mapping.allValues;
	
	// Check finder items
	NSURL *lastItem = nil;
	NSArray *finderItems = [[NSWorkspace sharedWorkspace] finderSidebarItems];
	for (NSURL *item in finderItems) {
		if ([deleted containsObject:item.path])
			[[NSWorkspace sharedWorkspace] removeFinderSidebarItem:item];
		if ([mappingPaths containsObject:item.path])
			lastItem = item;
	}
	
	for (NSString *path in created) {
		NSURL *item = [NSURL fileURLWithPath:path isDirectory:YES];
		if (![finderItems containsObject:item]) {
			[[NSWorkspace sharedWorkspace] insertFinderSidebarItem:item afterItem:lastItem];
			lastItem = item;
		}
	}
}

#pragma mark - Public

+ (void)setWorkspace:(NSString *)workspace {
	sharedInstance = nil;
	if (workspace) {
		sharedInstance = [[P4WorkspaceDefaults alloc] initWithWorkspace:workspace];
		// Save last workspace details
		P4Defaults *defaults = [P4Defaults sharedInstance];
		NSMutableDictionary *users = [NSMutableDictionary dictionaryWithDictionary:defaults.users];
		[users setObject:workspace forKey:[P4Workspace sharedInstance].username];
		defaults.users = users;
		defaults.workspace = workspace;
	}
}

- (void)updateMapping {
	
	NSDictionary *prevMapping = self.mapping;
	NSArray *prevPaths = [prevMapping allValues];
	
	// Prepare new mapping
	NSMutableArray *newPaths = [NSMutableArray array];
	NSMutableDictionary *newMapping = [NSMutableDictionary dictionary];
	[[[P4Workspace sharedInstance] mapping] enumerateKeysAndObjectsUsingBlock:^(id key, id obj, BOOL *stop) {
		if ([key hasPrefix:@"-"])
			return;
		NSString *path = [obj stringByRemovingSuffix:@"/"];
		[newPaths addObject:path];
		[newMapping setObject:path forKey:key];
	}];
	
	// Check for changes in mapping
	NSMutableArray *deleted = [prevPaths mutableCopy];
	[deleted removeObjectsInArray:newPaths];
	NSMutableArray *created = [newPaths mutableCopy];
	[created removeObjectsInArray:prevPaths];
	
	// Update symlinks and finder sidebar
	if (created.count || deleted.count) {
		[self updateSymbolicLinks:created deleted:deleted];
//		[self updateFinderSidebarItems:created deleted:deleted]; // Optional making finder items for every mapped folder
	}
	
	self.mapping = newMapping;
}

- (void)addFavoriteFolder:(NSString *)folder {
	NSMutableArray *mutableFolders = [NSMutableArray arrayWithArray:favoriteFolders];
	[mutableFolders addObject:folder];
	self.favoriteFolders = mutableFolders;
	
	[[NSNotificationCenter defaultCenter]
	 postNotificationName:P4WorkspaceDefaultsChangedNotification
	 object:folder];
}

- (void)removeFavoriteFolder:(NSString *)folder {
	NSMutableArray *mutableFolders = [NSMutableArray arrayWithArray:favoriteFolders];
	[mutableFolders removeObject:folder];
	self.favoriteFolders = mutableFolders;
	
	if (favoriteNames) {
		NSMutableDictionary *names = [NSMutableDictionary dictionaryWithDictionary:favoriteNames];
		[names removeObjectForKey:folder];
		self.favoriteNames = names;
	}
	
	[[NSNotificationCenter defaultCenter]
	 postNotificationName:P4WorkspaceDefaultsChangedNotification
	 object:folder];
}

- (void)renameFavoriteFolder:(NSString *)folder name:(NSString *)name {
	NSMutableDictionary *mutableNames = [NSMutableDictionary dictionaryWithDictionary:favoriteNames];
	[mutableNames setObject:name forKey:folder];
	self.favoriteNames = mutableNames;
}

- (void)removeFavoriteTag:(NSString *)tag {
	NSMutableArray *mutableTags = [NSMutableArray arrayWithArray:favoriteTags];
	[mutableTags removeObject:tag];
	self.favoriteTags = mutableTags;
	
	if ([filteredTags containsObject:tag]) {
		[self setFilteredTag:tag];
	}
}

- (void)setFilteredTag:(NSString *)tag {
	NSMutableArray *filtered = [NSMutableArray arrayWithArray:filteredTags];
	if ([filtered containsObject:tag])
		[filtered removeObject:tag];
	else
		[filtered addObject:tag];
	self.filteredTags = filtered;
	
	[[NSNotificationCenter defaultCenter]
	 postNotificationName:P4WorkspaceDefaultsTagFilterNotification
	 object:tag];
}

@end
